@@ -41,18 +41,31 @@ macro_rules! define_keywords {
     ($(
         $ident:ident $(= $string_keyword:expr)?
     ),*) => {
-        $(kw_def!($ident $(= $string_keyword)?);)*
+        #[derive(Debug, Clone, Copy, PartialEq, PartialOrd, Eq, Ord)]
+        #[allow(non_camel_case_types)]
+        pub enum Keyword {
+            NoKeyword,
+            $($ident),*
+        }
 
+        pub const ALL_KEYWORDS_INDEX: &[Keyword] = &[
+            $(Keyword::$ident),*
+        ];
+
+        $(kw_def!($ident $(= $string_keyword)?);)*
         pub const ALL_KEYWORDS: &[&str] = &[
             $($ident),*
         ];
-    }
+
+    };
+
 }
 
+// The following keywords should be sorted to be able to match using binary search
 define_keywords!(
     ABS,
+    ACTION,
     ADD,
-    ASC,
     ALL,
     ALLOCATE,
     ALTER,
@@ -64,12 +77,14 @@ define_keywords!(
     ARRAY_AGG,
     ARRAY_MAX_CARDINALITY,
     AS,
+    ASC,
     ASENSITIVE,
     ASYMMETRIC,
     AT,
     ATOMIC,
     AUTHORIZATION,
     AVG,
+    AVRO,
     BEGIN,
     BEGIN_FRAME,
     BEGIN_PARTITION,
@@ -92,9 +107,9 @@ define_keywords!(
     CEILING,
     CHAIN,
     CHAR,
-    CHAR_LENGTH,
     CHARACTER,
     CHARACTER_LENGTH,
+    CHAR_LENGTH,
     CHECK,
     CLOB,
     CLOSE,
@@ -102,6 +117,7 @@ define_keywords!(
     COLLATE,
     COLLECT,
     COLUMN,
+    COLUMNS,
     COMMIT,
     COMMITTED,
     CONDITION,
@@ -156,9 +172,11 @@ define_keywords!(
     ELEMENT,
     ELSE,
     END,
+    END_EXEC = "END-EXEC",
     END_FRAME,
     END_PARTITION,
     EQUALS,
+    ERROR,
     ESCAPE,
     EVERY,
     EXCEPT,
@@ -166,12 +184,14 @@ define_keywords!(
     EXECUTE,
     EXISTS,
     EXP,
+    EXTENDED,
     EXTERNAL,
     EXTRACT,
     FALSE,
     FETCH,
-    FIRST,
+    FIELDS,
     FILTER,
+    FIRST,
     FIRST_VALUE,
     FLOAT,
     FLOOR,
@@ -197,6 +217,7 @@ define_keywords!(
     IDENTITY,
     IF,
     IN,
+    INDEX,
     INDICATOR,
     INNER,
     INOUT,
@@ -211,10 +232,12 @@ define_keywords!(
     IS,
     ISOLATION,
     JOIN,
+    JSONFILE,
     KEY,
     LAG,
     LANGUAGE,
     LARGE,
+    LAST,
     LAST_VALUE,
     LATERAL,
     LEAD,
@@ -224,6 +247,7 @@ define_keywords!(
     LIKE,
     LIKE_REGEX,
     LIMIT,
+    LISTAGG,
     LN,
     LOCAL,
     LOCALTIME,
@@ -247,8 +271,8 @@ define_keywords!(
     NATURAL,
     NCHAR,
     NCLOB,
-    NEXT,
     NEW,
+    NEXT,
     NO,
     NONE,
     NORMALIZE,
@@ -257,10 +281,11 @@ define_keywords!(
     NTILE,
     NULL,
     NULLIF,
+    NULLS,
     NUMERIC,
     OBJECT,
-    OCTET_LENGTH,
     OCCURRENCES_REGEX,
+    OCTET_LENGTH,
     OF,
     OFFSET,
     OLD,
@@ -268,19 +293,21 @@ define_keywords!(
     ONLY,
     OPEN,
     OR,
+    ORC,
     ORDER,
     OUT,
     OUTER,
     OVER,
+    OVERFLOW,
     OVERLAPS,
     OVERLAY,
     PARAMETER,
-    PARTITION,
     PARQUET,
+    PARTITION,
     PERCENT,
-    PERCENT_RANK,
     PERCENTILE_CONT,
     PERCENTILE_DISC,
+    PERCENT_RANK,
     PERIOD,
     PORTION,
     POSITION,
@@ -294,6 +321,7 @@ define_keywords!(
     PROCEDURE,
     RANGE,
     RANK,
+    RCFILE,
     READ,
     READS,
     REAL,
@@ -312,6 +340,7 @@ define_keywords!(
     REGR_SXY,
     REGR_SYY,
     RELEASE,
+    RENAME,
     REPEATABLE,
     RESTRICT,
     RESULT,
@@ -322,18 +351,22 @@ define_keywords!(
     ROLLBACK,
     ROLLUP,
     ROW,
-    ROW_NUMBER,
     ROWS,
+    ROW_NUMBER,
     SAVEPOINT,
+    SCHEMA,
     SCOPE,
     SCROLL,
     SEARCH,
     SECOND,
     SELECT,
     SENSITIVE,
+    SEQUENCEFILE,
     SERIALIZABLE,
+    SESSION,
     SESSION_USER,
     SET,
+    SHOW,
     SIMILAR,
     SMALLINT,
     SOME,
@@ -362,6 +395,7 @@ define_keywords!(
     TABLE,
     TABLESAMPLE,
     TEXT,
+    TEXTFILE,
     THEN,
     TIES,
     TIME,
@@ -369,6 +403,7 @@ define_keywords!(
     TIMEZONE_HOUR,
     TIMEZONE_MINUTE,
     TO,
+    TOP,
     TRAILING,
     TRANSACTION,
     TRANSLATE,
@@ -376,10 +411,10 @@ define_keywords!(
     TRANSLATION,
     TREAT,
     TRIGGER,
-    TRUNCATE,
     TRIM,
     TRIM_ARRAY,
     TRUE,
+    TRUNCATE,
     UESCAPE,
     UNBOUNDED,
     UNCOMMITTED,
@@ -395,11 +430,11 @@ define_keywords!(
     VALUE,
     VALUES,
     VALUE_OF,
-    VAR_POP,
-    VAR_SAMP,
     VARBINARY,
     VARCHAR,
     VARYING,
+    VAR_POP,
+    VAR_SAMP,
     VERSIONING,
     VIEW,
     WHEN,
@@ -410,29 +445,60 @@ define_keywords!(
     WITH,
     WITHIN,
     WITHOUT,
-    WRITE,
     WORK,
+    WRITE,
     YEAR,
-    ZONE,
-    END_EXEC = "END-EXEC"
+    ZONE
 );
 
 /// These keywords can't be used as a table alias, so that `FROM table_name alias`
 /// can be parsed unambiguously without looking ahead.
-pub const RESERVED_FOR_TABLE_ALIAS: &[&str] = &[
+pub const RESERVED_FOR_TABLE_ALIAS: &[Keyword] = &[
     // Reserved as both a table and a column alias:
-    WITH, SELECT, WHERE, GROUP, HAVING, ORDER, LIMIT, OFFSET, FETCH, UNION, EXCEPT, INTERSECT,
+    Keyword::WITH,
+    Keyword::SELECT,
+    Keyword::WHERE,
+    Keyword::GROUP,
+    Keyword::HAVING,
+    Keyword::ORDER,
+    Keyword::TOP,
+    Keyword::LIMIT,
+    Keyword::OFFSET,
+    Keyword::FETCH,
+    Keyword::UNION,
+    Keyword::EXCEPT,
+    Keyword::INTERSECT,
     // Reserved only as a table alias in the `FROM`/`JOIN` clauses:
-    ON, JOIN, INNER, CROSS, FULL, LEFT, RIGHT, NATURAL, USING,
+    Keyword::ON,
+    Keyword::JOIN,
+    Keyword::INNER,
+    Keyword::CROSS,
+    Keyword::FULL,
+    Keyword::LEFT,
+    Keyword::RIGHT,
+    Keyword::NATURAL,
+    Keyword::USING,
     // for MSSQL-specific OUTER APPLY (seems reserved in most dialects)
-    OUTER,
+    Keyword::OUTER,
 ];
 
 /// Can't be used as a column alias, so that `SELECT <expr> alias`
 /// can be parsed unambiguously without looking ahead.
-pub const RESERVED_FOR_COLUMN_ALIAS: &[&str] = &[
+pub const RESERVED_FOR_COLUMN_ALIAS: &[Keyword] = &[
     // Reserved as both a table and a column alias:
-    WITH, SELECT, WHERE, GROUP, HAVING, ORDER, LIMIT, OFFSET, FETCH, UNION, EXCEPT, INTERSECT,
-    // Reserved only as a column alias in the `SELECT` clause:
-    FROM,
+    Keyword::WITH,
+    Keyword::SELECT,
+    Keyword::WHERE,
+    Keyword::GROUP,
+    Keyword::HAVING,
+    Keyword::ORDER,
+    Keyword::TOP,
+    Keyword::LIMIT,
+    Keyword::OFFSET,
+    Keyword::FETCH,
+    Keyword::UNION,
+    Keyword::EXCEPT,
+    Keyword::INTERSECT,
+    // Reserved only as a column alias in the `SELECT` clause
+    Keyword::FROM,
 ];
